#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Finalizations. Last steps directly before code creation is called.

Here the final tasks are executed. Things normally volatile during optimization
can be computed here, so the code generation can be quick and doesn't have to
check it many times.

"""
from .FinalizeMarkups import FinalizeMarkups
from .FinalizeClosureTaking import FinalizeClosureTaking

# Bug of pylint, it's there but it reports it wrongly, pylint: disable=E0611
from nuitka.tree import Operations

def prepareCodeGeneration( tree ):
    visitor = FinalizeMarkups()
    Operations.visitTree( tree, visitor )
    for function in tree.getUsedFunctions():
        Operations.visitTree( function, visitor )

    visitor = FinalizeClosureTaking()
    for function in tree.getUsedFunctions():
        Operations.visitFunction( function, visitor )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Scons interface.

Interaction with scons. Find the binary, and run it with a set of given
options.

"""

from nuitka import Options, Tracing, Utils

import os, sys

def getSconsDataPath():
    return Utils.dirname( __file__ )

def getSconsInlinePath():
    return Utils.joinpath( getSconsDataPath(), "inline_copy" )

def getSconsBinaryPath():
    """ Return a way to execute Scons.

        Using potentially inline copy if no system Scons is available
        or if we are on Windows.
    """

    if Utils.isFile( "/usr/bin/scons" ):
        return "/usr/bin/scons"
    else:
        return "%s %s" % (
            getPython2ExePath(),
            Utils.joinpath( getSconsInlinePath(), "bin", "scons.py" )
        )

def getPython2ExePath():
    """ Find a way to call Python2. Scons needs it."""

    if Utils.python_version < 300:
        return sys.executable
    elif os.name == "nt":
        if os.path.exists( r"c:\Python27\python.exe" ):
            return r"c:\Python27\python.exe"
        elif os.path.exists( r"c:\Python26\python.exe" ):
            return r"c:\Python26\python.exe"
        else:
            sys.exit( """\
Error, need to find Python2 executable under C:\\Python26 or \
C:\\Python27 to execute scons which is not Python3 compatible.""" )
    elif os.path.exists( "/usr/bin/python2" ):
        return "python2"
    else:
        return "python"


def runScons( options, quiet ):
    # For the scons file to find the static C++ files and include path. The
    # scons file is unable to use __file__ for the task.
    os.environ[ "NUITKA_SCONS" ] = getSconsDataPath()

    if os.name == "nt":
        # On Windows this Scons variable must be set by us.
        os.environ[ "SCONS_LIB_DIR" ] = Utils.joinpath(
            getSconsInlinePath(),
            "lib",
            "scons-2.0.1"
        )

        # Also, for MinGW we can avoid the user having to add the path if he
        # used the default path or installed it on the same drive by appending
        # to the PATH variable before executing scons.
        os.environ[ "PATH" ] += r";\MinGW\bin;C:\MinGW\bin"

    # Scons is Python2 only, so we need to make the system find a suitable
    # Python binary.
    python2_exe = getPython2ExePath()

    scons_command = """%(scons_call)s %(quiet)s --warn=no-deprecated \
-f %(scons_file)s --jobs %(job_limit)d %(options)s""" % {
        "scons_call" : getSconsBinaryPath(),
        "quiet"      : "--quiet" if quiet else "",
        "scons_file" : Utils.joinpath( getSconsDataPath(), "SingleExe.scons" ),
        "job_limit"  : Options.getJobLimit(),
        "options"    : " ".join(
            "%s=%s" % ( key, value )
            for key, value in
            options.items()
        )
    }

    if Options.isShowScons():
        Tracing.printLine( "Scons command:", scons_command )

    return 0 == os.system( scons_command )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Identifiers hold code references.

These are generally the means to effectively hide the reference count. The best part
is where getCheapRefCount tries to not allocate references not needed.
"""

# The method signatures do not always require usage of self, sometimes can be decided
# based on class. pylint: disable=R0201

from nuitka import Utils

def encodeNonAscii( var_name ):
    if Utils.python_version < 300:
        return var_name
    else:
        var_name = var_name.encode( "ascii", "xmlcharrefreplace" )

        return var_name.decode( "ascii" ).replace( "&#", "$$" ).replace( ";", "" )

class Identifier:
    def __init__( self, code, ref_count ):
        self.code = code
        self.ref_count = ref_count

    def getRefCount( self ):
        return self.ref_count

    def getCheapRefCount( self ):
        return self.ref_count

    def getCode( self ):
        return self.code

    def getCodeObject( self ):
        return self.code

    def getCodeExportRef( self ):
        if self.getRefCount():
            return self.getCodeObject()
        else:
            return "INCREASE_REFCOUNT( %s )" % self.getCodeObject()

    def getCodeTemporaryRef( self ):
        if self.getRefCount():
            return "PyObjectTemporary( %s ).asObject()" % self.getCodeObject()
        else:
            return self.getCodeObject()

    def getCodeDropRef( self ):
        if self.ref_count == 0:
            return self.getCodeTemporaryRef()
        else:
            return "DECREASE_REFCOUNT( %s )" % self.getCodeObject()

    def __repr__( self ):
        return "<Identifier %s (%d)>" % ( self.code, self.ref_count )

    def isConstantIdentifier( self ):
        return False


class ConstantIdentifier( Identifier ):
    def __init__( self, constant_code, constant_value ):
        Identifier.__init__( self, constant_code, 0 )

        self.constant_value = constant_value

    def __repr__( self ):
        return "<ConstantIdentifier %s>" % self.code

    def isConstantIdentifier( self ):
        return True

    def getCheapRefCount( self ):
        return 0

    def getConstant( self ):
        return self.constant_value


class SpecialConstantIdentifier( ConstantIdentifier ):
    def __init__( self, constant_value ):
        if constant_value is None:
            ConstantIdentifier.__init__( self, "Py_None", None )
        elif constant_value is True:
            ConstantIdentifier.__init__( self, "Py_True", True )
        elif constant_value is False:
            ConstantIdentifier.__init__( self, "Py_False", False )
        elif constant_value is Ellipsis:
            ConstantIdentifier.__init__( self, "Py_Ellipsis", Ellipsis )
        else:
            assert False, constant_value


class EmptyDictIdentifier( Identifier ):
    def __init__( self ):
        Identifier.__init__( self, "PyDict_New()", 1 )

    def getCheapRefCount( self ):
        return 1

    def isConstantIdentifier( self ):
        return True

    def getConstant( self ):
        return {}


class ModuleVariableIdentifier:
    def __init__( self, var_name, module_code_name ):
        self.var_name = var_name
        self.module_code_name = module_code_name

    def isConstantIdentifier( self ):
        return False

    def __repr__( self ):
        return "<ModuleVariableIdentifier %s>" % self.var_name

    def getRefCount( self ):
        return 0

    def getCheapRefCount( self ):
        # The asObject0 is the fastest way, stealing a reference directly from the module
        # dictionary if possible.
        return 0

    def getCodeTemporaryRef( self ):
        return "_mvar_%s_%s.asObject0()" % ( self.module_code_name, self.var_name )

    def getCodeExportRef( self ):
        return "_mvar_%s_%s.asObject()" % ( self.module_code_name, self.var_name )

    def getCodeDropRef( self ):
        return self.getCodeTemporaryRef()

    def getCode( self ):
        return "_mvar_%s_%s" % ( self.module_code_name, encodeNonAscii( self.var_name ) )


class MaybeModuleVariableIdentifier( Identifier ):
    def __init__( self, var_name, module_code_name ):
        Identifier.__init__(
            self,
            "_mvar_%s_%s.asObject0( locals_dict.asObject() )" % (
                module_code_name,
                var_name
            ),
            0
        )


class LocalVariableIdentifier:
    def __init__( self, var_name, from_context = False ):
        assert type( var_name ) == str

        self.from_context = from_context
        self.var_name = var_name

    def isConstantIdentifier( self ):
        return False

    def __repr__( self ):
        return "<LocalVariableIdentifier %s>" % self.var_name

    def getCode( self ):
        if not self.from_context:
            return "_python_var_" + encodeNonAscii( self.var_name )
        else:
            return "_python_context->python_var_" + encodeNonAscii( self.var_name )

    def getRefCount( self ):
        return 0

    def getCheapRefCount( self ):
        return 0

    def getCodeObject( self ):
        return "%s.asObject()" % self.getCode()

    def getCodeTemporaryRef( self ):
        return "%s.asObject()" % self.getCode()

    def getCodeExportRef( self ):
        return "%s.asObject1()" % self.getCode()

    def getCodeDropRef( self ):
        return self.getCodeTemporaryRef()


class TempVariableIdentifier( Identifier ):
    def __init__( self, var_name ):
        self.tempvar_name = var_name

        Identifier.__init__( self, "_python_tmp_" + var_name, 0 )

    def __repr__( self ):
        return "<TempVariableIdentifier %s >" % self.tempvar_name

    def getCheapRefCount( self ):
        return 0

    def getCodeObject( self ):
        return "%s.asObject()" % self.getCode()

    def getClass( self ):
        return "PyObjectTemporary"


class TempObjectIdentifier( Identifier ):
    def __init__( self, var_name ):
        self.tempvar_name = var_name

        Identifier.__init__( self, "_python_tmp_" + var_name, 0 )

    def getCodeTemporaryRef( self ):
        return self.code



class KeeperAccessIdentifier( Identifier ):
    def __init__( self, var_name ):
        Identifier.__init__( self, var_name, 1 )

    def getCheapRefCount( self ):
        return 0


class ClosureVariableIdentifier( Identifier ):
    def __init__( self, var_name, from_context ):
        assert type( from_context ) is str

        self.var_name = var_name
        self.from_context = from_context

        if self.from_context:
            Identifier.__init__( self, self.from_context + "python_closure_" + encodeNonAscii( self.var_name ), 0 )
        else:
            # TODO: Use a variable object to decide naming policy

            Identifier.__init__( self, "python_closure_" + encodeNonAscii( self.var_name ), 0 )

    def __repr__( self ):
        return "<ClosureVariableIdentifier %s >" % self.var_name

    def getCheapRefCount( self ):
        return 0

    def getCodeObject( self ):
        return self.getCode() + ".asObject()"


class NullIdentifier( Identifier ):
    def __init__( self ):
        Identifier.__init__(
            self,
            code      = "NULL",
            ref_count = 0
        )

    def getCodeExportRef( self ):
        return "NULL"


class ThrowingIdentifier( Identifier ):
    def __init__( self, code ):
        Identifier.__init__(
            self,
            code      = code,
            ref_count = 0
        )

    def getCodeExportRef( self ):
        return self.getCodeObject()

    def getCodeTemporaryRef( self ):
        return self.getCodeObject()

    def getCheapRefCount( self ):
        return 0


class CallIdentifier( Identifier ):
    def __init__( self, called, args ):
        Identifier.__init__(
            self,
            code      = "%s( %s )" % (
                called,
                ", ".join( args )
            ),
            ref_count = 1
        )


class HelperCallIdentifier( CallIdentifier ):
    def __init__( self, helper, *args ):
        CallIdentifier.__init__(
            self,
            called = helper,
            args   = [
                arg.getCodeTemporaryRef() if arg is not None else "NULL"
                for arg in
                args
            ]
        )


def getCodeTemporaryRefs( identifiers ):
    """ Helper to create temporary reference code of many identifiers at once.

    """

    return [ identifier.getCodeTemporaryRef() for identifier in identifiers ]

def getCodeExportRefs( identifiers ):
    """ Helper to create export reference code of many identifiers at once.

    """

    return [ identifier.getCodeExportRef() for identifier in identifiers ]



#!/usr/bin/env python
#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from __future__ import print_function

import os, sys, subprocess, tempfile, shutil

# Go its own directory, to have it easy with path knowledge.
os.chdir( os.path.dirname( os.path.abspath( __file__ ) ) )

search_mode = len( sys.argv ) > 1 and sys.argv[1] == "search"

start_at = sys.argv[2] if len( sys.argv ) > 2 else None

if start_at:
    active = False
else:
    active = True

if "PYTHON" not in os.environ:
    os.environ[ "PYTHON" ] = sys.executable

def check_output(*popenargs, **kwargs):
    from subprocess import Popen, PIPE, CalledProcessError

    if 'stdout' in kwargs:
        raise ValueError('stdout argument not allowed, it will be overridden.')
    process = Popen(stdout=PIPE, *popenargs, **kwargs)
    output, unused_err = process.communicate()
    retcode = process.poll()
    if retcode:
        cmd = kwargs.get("args")
        if cmd is None:
            cmd = popenargs[0]
        raise CalledProcessError(retcode, cmd, output=output)
    return output

version_output = check_output(
    [ os.environ[ "PYTHON" ], "--version" ],
    stderr = subprocess.STDOUT
)

python_version = version_output.split()[1]

os.environ[ "PYTHONPATH" ] = os.getcwd()

print( "Using concrete python", python_version )

for filename in sorted( os.listdir( "." ) ):
    if not filename.endswith( ".py" ) or filename == "run_all.py":
        continue

    path = filename

    if not active and start_at in ( filename, path ):
        active = True

    # Some syntax errors are for Python3 only.
    if filename == "Importing2.py" and python_version < "3":
        extra_flags = [ "remove_output" ]
    elif filename == "GeneratorReturn2.py" and python_version >= "3.3":
        extra_flags = [ "remove_output" ]
    else:
        extra_flags = [ "expect_failure",  "remove_output" ]

    if active:
        result = subprocess.call(
            "%s %s %s silent %s" % (
                sys.executable,
                os.path.join( "..", "..", "bin", "compare_with_cpython" ),
                path,
                " ".join( extra_flags )
            ),
            shell = True
        )

        if result == 2:
            sys.stderr.write( "Interruped, with CTRL-C\n" )
            sys.exit( 2 )

        if result != 0 and search_mode:
            print( "Error exit!", result )
            sys.exit( result )
    else:
        print( "Skipping", filename )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Pack and copy files for portable mode.

usage: PortableSetup.py mainscript outputdir

"""

import sys
import os
import zipfile

python_library_archive_name = "_python.zip"
python_dll_dir_name = "_python"

python_executable_suffixes = ( ".py", ".pyc", ".pyo" )
builtin_module_names = list( sys.builtin_module_names ) + [ "_io" ]

def importList( *names ):
    for name in names:
        __import__( name )

dependency_resolver = {
    "_ssl": ( importList, "socket", "_socket" ),
    "PyQt4.QtGui" : ( importList, "sip" ),
}

def isPythonScript( path ):
    for end in python_executable_suffixes:
        if path.endswith( end ):
            return 1
    return 0

def copyFile( src, dst ):
    target_dir = os.path.dirname( dst )

    if not os.path.isdir( target_dir ):
        os.makedirs( target_dir )

    if os.path.isfile( dst ):
        os.remove( dst )
    with open( src, "rb" ) as p:
        data = p.read()
    with open( dst, "wb" ) as p:
        p.write( data )
        p.flush()

def getImportedDict( mainscript ):
    # importing a lot of stuff, just because they are dependencies.
    # pylint: disable=W0612,R0914

    # chdir to mainscript directory and add it to sys.path
    main_dir = os.path.dirname( mainscript )
    os.chdir( main_dir )
    sys.path.insert( 0, main_dir )

    # import modules needed but not listed in sys.modules
    import imp, zipimport, site, io, marshal, pickle
    import encodings, encodings.aliases, codecs
    import zlib, inspect, threading, traceback
    import ctypes
    if sys.version_info < ( 3, 0, 0 ):
        import StringIO, cStringIO
        import cPickle
        import thread
    for code in encodings.aliases.aliases.keys():
        try:
            encodings.search_function( code )
        except ( ImportError, AttributeError ):
            pass

    # get modules from main script
    import modulefinder
    finder = modulefinder.ModuleFinder()
    finder.run_script( mainscript )

    # resolve dependency
    for name in finder.modules:
        method = dependency_resolver.get( name )
        if not method:
            continue
        method[0]( *method[1:] )

    # merge sys.modules
    imported_dict = {}
    imported_dict.update( sys.modules )
    imported_dict.update( finder.modules )
    return imported_dict

def yieldImportedPaths( imported_dict ):
    module_base = os.path.dirname( os.__file__ )
    if os.name == "nt":
        module_base_dlls = os.path.join(
            os.path.dirname( sys.executable ),
            "DLLs"
        )
    for name, mod in imported_dict.items():
        if not mod:
            continue

        if not hasattr( mod, "__file__" ) or mod.__file__ is None:
            # builtin module, if mod is modulefinder.Module then __file__ will
            # be None
            if name not in builtin_module_names:
                import warnings
                warnings.warn( "unknown builtin module %s\n" % repr( name ), Warning )
            continue

        path = mod.__file__
        if path.startswith( module_base ):
            yield name, path
        elif os.name == "nt" and not isPythonScript( path ) and \
           path.startswith( module_base_dlls ):
            yield name, path
            continue

def copyPythonLibrary( outputdir ):
    if os.name == "posix" and os.uname()[0] == "Linux":
        with open( "/proc/%s/smaps" % os.getpid() ) as pmap:
            for line in pmap:
                if line.find("libpython") != -1:
                    src = line[ line.find( "/" ): ].strip()
                    dst = os.path.join( outputdir, os.path.basename ( src ) )
                    copyFile( src, dst )
                    break
    elif os.name == "nt":
        import ctypes
        from ctypes import windll
        from ctypes.wintypes import HANDLE, LPCSTR, DWORD
        dll = getattr( windll, "python%s%s" % ( sys.version_info[:2] ) )
        getname = windll.kernel32.GetModuleFileNameA
        getname.argtypes = ( HANDLE, LPCSTR, DWORD )
        getname.restype = DWORD
        result = ctypes.create_string_buffer( 1024 )
        size = getname( dll._handle, result, 1024 ) # needed for this call, pylint: disable=W0212
        src = result.value[ :size ]
        dst = os.path.join( outputdir, os.path.basename ( src ) )
        copyFile( src, dst )
    else:
        # TODO: Add support for bsd and osx here
        sys.exit( "Error, unsupported platform for portable binaries." )

def main( mainscript, outputdir ):
    imported_dict = getImportedDict( mainscript )

    zip_list = []
    bin_list = []

    for name, path in yieldImportedPaths( imported_dict ):
        if isPythonScript( path ):
            path_base = path.rsplit( ".", 1 )[0]
            for end in python_executable_suffixes:
                path_pack = path_base + end
                if os.path.isfile( path_pack ):
                    if path_base.endswith( "__init__" ):
                        import_base = os.path.dirname( path_base )
                    else:
                        import_base = path_base

                    for _i in range( name.count(".") + 1 ):
                        import_base = os.path.dirname( import_base )

                    zip_list.append( ( path_pack, import_base ) )
                    break
        else:
            import_base = path
            for _i in range( name.count(".") + 1 ):
                import_base = os.path.dirname( import_base )

            export_path = path[ len( import_base ) + 1 : ]

            bin_list.append( ( path, export_path ) )

    # pack scripts to archive
    zip_path = os.path.join( outputdir, python_library_archive_name )
    with zipfile.ZipFile( zip_path, "w", zipfile.ZIP_STORED ) as zip_file:
        for path_pack, import_base in sorted( zip_list ):
            zip_file.write( path_pack, path_pack[ len( import_base ): ] )

    # copy extensions to directory
    import shutil
    library_directory = os.path.join( outputdir, python_dll_dir_name )
    if os.path.isdir( library_directory ):
        shutil.rmtree( library_directory )
    if not os.path.isdir( library_directory ):
        os.makedirs( library_directory )
    for src, export_path in bin_list:
        dst = os.path.join( library_directory, export_path )
        copyFile( src, dst )

    # copy libpython
    copyPythonLibrary( outputdir )

def setup( mainscript, outputdir ):
    # if use this script as module, use this method
    import subprocess
    proc = subprocess.Popen(
        args   = ( sys.executable, __file__, mainscript, outputdir ),
        stdout = sys.stdout,
        stderr = sys.stderr,
        stdin  = sys.stdin,
        shell  = 0
    )
    proc.wait()
    return proc.poll() == 0

if __name__ == "__main__":
    main(
        mainscript = os.path.abspath( sys.argv[1] ),
        outputdir  = os.path.abspath( sys.argv[2] )
    )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from nuitka.nodes.VariableRefNodes import (
    ExpressionTargetTempVariableRef,
    ExpressionTempVariableRef,
    StatementTempBlock
)
from nuitka.nodes.ConstantRefNodes import ExpressionConstantRef

from nuitka.nodes.BuiltinIteratorNodes import (
    ExpressionBuiltinNext1,
    ExpressionBuiltinIter1
)
from nuitka.nodes.ComparisonNodes import ExpressionComparisonIs
from nuitka.nodes.StatementNodes import StatementsSequence
from nuitka.nodes.LoopNodes import (
    StatementBreakLoop,
    StatementLoop
)
from nuitka.nodes.ConditionalNodes import StatementConditional
from nuitka.nodes.AssignNodes import StatementAssignmentVariable

from .Helpers import (
    makeStatementsSequenceFromStatement,
    makeStatementsSequence,
    buildStatementsNode,
    buildNode
)

from .ReformulationAssignmentStatements import buildAssignmentStatements
from .ReformulationTryExceptStatements import makeTryExceptSingleHandlerNode


def buildForLoopNode( provider, node, source_ref ):
    # The for loop is re-formulated according to developer manual. An iterator is created,
    # and looped until it gives StopIteration. The else block is taken if a for loop exits
    # normally, i.e. because of iterator exhaustion. We do this by introducing an
    # indicator variable.

    source = buildNode( provider, node.iter, source_ref )

    result = StatementTempBlock(
        source_ref = source_ref
    )

    tmp_iter_variable = result.getTempVariable( "for_iterator" )

    iterate_tmp_block = StatementTempBlock(
        source_ref = source_ref
    )

    tmp_value_variable = iterate_tmp_block.getTempVariable( "iter_value" )

    else_block = buildStatementsNode(
        provider   = provider,
        nodes      = node.orelse if node.orelse else None,
        source_ref = source_ref
    )

    if else_block is not None:
        tmp_break_indicator_variable = result.getTempVariable( "break_indicator" )

        statements = [
            StatementAssignmentVariable(
                variable_ref = ExpressionTargetTempVariableRef(
                    variable   = tmp_break_indicator_variable.makeReference( result ),
                    source_ref = source_ref
                ),
                source     = ExpressionConstantRef(
                    constant   = True,
                    source_ref = source_ref
                ),
                source_ref = source_ref
            )
        ]
    else:
        statements = []

    statements.append(
        StatementBreakLoop(
            source_ref = source_ref.atInternal()
        )
    )

    handler_body = makeStatementsSequence(
        statements = statements,
        allow_none = False,
        source_ref = source_ref
    )

    statements = (
        makeTryExceptSingleHandlerNode(
            tried         = makeStatementsSequenceFromStatement(
                statement = StatementAssignmentVariable(
                    variable_ref = ExpressionTargetTempVariableRef(
                        variable   = tmp_value_variable.makeReference( iterate_tmp_block ),
                        source_ref = source_ref
                    ),
                    source     = ExpressionBuiltinNext1(
                        value      = ExpressionTempVariableRef(
                            variable   = tmp_iter_variable.makeReference( result ),
                            source_ref = source_ref
                        ),
                        source_ref = source_ref
                    ),
                    source_ref = source_ref
                )
            ),
            exception_name = "StopIteration",
            handler_body   = handler_body,
            source_ref     = source_ref
        ),
        buildAssignmentStatements(
            provider   = provider,
            node       = node.target,
            source     = ExpressionTempVariableRef(
                variable   = tmp_value_variable.makeReference( iterate_tmp_block ),
                source_ref = source_ref
            ),
            source_ref = source_ref
        )
    )

    iterate_tmp_block.setBody(
        StatementsSequence(
            statements = statements,
            source_ref = source_ref
        )
    )

    statements = (
        iterate_tmp_block,
        buildStatementsNode(
            provider   = provider,
            nodes      = node.body,
            source_ref = source_ref
        )
    )

    loop_body = makeStatementsSequence(
        statements = statements,
        allow_none = True,
        source_ref = source_ref
    )

    if else_block is not None:
        statements = [
            StatementAssignmentVariable(
                variable_ref = ExpressionTargetTempVariableRef(
                    variable   = tmp_break_indicator_variable.makeReference( result ),
                    source_ref = source_ref
                ),
                source     = ExpressionConstantRef(
                    constant = False,
                    source_ref = source_ref
                ),
                source_ref = source_ref
            )
        ]
    else:
        statements = []

    statements += [
        # First create the iterator and store it.
        StatementAssignmentVariable(
            variable_ref = ExpressionTargetTempVariableRef(
                variable   = tmp_iter_variable.makeReference( result ),
                source_ref = source_ref
            ),
            source     = ExpressionBuiltinIter1(
                value       = source,
                source_ref  = source.getSourceReference()
            ),
            source_ref = source_ref
        ),
        StatementLoop(
            body       = loop_body,
            source_ref = source_ref
        )
    ]

    if else_block is not None:
        statements += [
            StatementConditional(
                condition  = ExpressionComparisonIs(
                    left       = ExpressionTempVariableRef(
                        variable   = tmp_break_indicator_variable.makeReference( result ),
                        source_ref = source_ref
                    ),
                    right      = ExpressionConstantRef(
                        constant   = True,
                        source_ref = source_ref
                    ),
                    source_ref = source_ref
                ),
                yes_branch = else_block,
                no_branch  = None,
                source_ref = source_ref
            )
        ]

    result.setBody(
        StatementsSequence(
            statements = statements,
            source_ref = source_ref
        )
    )

    return result

def buildWhileLoopNode( provider, node, source_ref ):
    # The while loop is re-formulated according to developer manual. The condition becomes
    # an early condition to break the loop. The else block is taken if a while loop exits
    # normally, i.e. because of condition not being true. We do this by introducing an
    # indicator variable.

    else_block = buildStatementsNode(
        provider   = provider,
        nodes      = node.orelse if node.orelse else None,
        source_ref = source_ref
    )

    if else_block is not None:
        temp_block = StatementTempBlock(
            source_ref = source_ref
        )

        tmp_break_indicator_variable = temp_block.getTempVariable( "break_indicator" )

        statements = (
            StatementAssignmentVariable(
                variable_ref = ExpressionTargetTempVariableRef(
                    variable   = tmp_break_indicator_variable.makeReference( temp_block ),
                    source_ref = source_ref
                ),
                source     = ExpressionConstantRef(
                    constant   = True,
                    source_ref = source_ref
                ),
                source_ref = source_ref
            ),
            StatementBreakLoop(
                source_ref = source_ref
            )
        )
    else:
        statements = (
            StatementBreakLoop(
                source_ref = source_ref
            ),
        )

    # The loop body contains a conditional statement at the start that breaks the loop if
    # it fails.
    loop_body = makeStatementsSequence(
        statements = (
            StatementConditional(
                condition = buildNode( provider, node.test, source_ref ),
                no_branch = StatementsSequence(
                    statements = statements,
                    source_ref = source_ref
                ),
                yes_branch = None,
                source_ref = source_ref
            ),
            buildStatementsNode(
                provider   = provider,
                nodes      = node.body,
                source_ref = source_ref
            )
        ),
        allow_none = True,
        source_ref = source_ref
    )

    loop_statement = StatementLoop(
        body       = loop_body,
        source_ref = source_ref
    )

    if else_block is None:
        return loop_statement
    else:
        statements = (
            StatementAssignmentVariable(
                variable_ref = ExpressionTargetTempVariableRef(
                    variable   = tmp_break_indicator_variable.makeReference( temp_block ),
                    source_ref = source_ref
                ),
                source = ExpressionConstantRef(
                    constant   = False,
                    source_ref = source_ref
                ),
                source_ref   = source_ref
            ),
            loop_statement,
            StatementConditional(
                condition  = ExpressionComparisonIs(
                    left       = ExpressionTempVariableRef(
                        variable   = tmp_break_indicator_variable.makeReference( temp_block ),
                        source_ref = source_ref
                    ),
                    right      = ExpressionConstantRef(
                        constant   = True,
                        source_ref = source_ref
                    ),
                    source_ref = source_ref
                ),
                yes_branch = else_block,
                no_branch  = None,
                source_ref = source_ref
            )
        )

        temp_block.setBody(
            StatementsSequence(
               statements = statements,
               source_ref = source_ref
            )
        )

        return temp_block



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Specification record for future flags.

A source reference also implies a specific set of future flags in use by the parser at
that location. Can be different inside a module due to e.g. the inlining of exec
statements with their own future imports, or inlining of code from other modules.
"""

from nuitka import Utils

_future_division_default = Utils.python_version >= 300
_future_absolute_import_default = Utils.python_version >= 300

class FutureSpec:
    def __init__( self ):
        self._future_division   = _future_division_default
        self._unicode_literals  = False
        self._absolute_import   = _future_absolute_import_default
        self._future_print      = False
        self._barry_bdfl        = False

    def clone( self ):
        result = FutureSpec()

        result._future_division   = self._future_division
        result._unicode_literals  = self._unicode_literals
        result._absolute_import   = self._absolute_import
        result._future_print      = self._future_print
        result._barry_bdfl        = self._barry_bdfl

        return result

    def isFutureDivision( self ):
        return self._future_division

    def enableFutureDivision( self ):
        self._future_division = True

    def enableFuturePrint( self ):
        self._future_print = True

    def enableUnicodeLiterals( self ):
        self._unicode_literals = True

    def enableAbsoluteImport( self ):
        self._absolute_import = True

    def enableBarry( self ):
        self._barry_bdfl = True

    def isAbsoluteImport( self ):
        return self._absolute_import

    def asFlags( self ):
        result = []

        if self._future_division:
            result.append( "CO_FUTURE_DIVISION" )

        if self._unicode_literals:
            result.append( "CO_FUTURE_UNICODE_LITERALS" )

        if self._absolute_import:
            result.append( "CO_FUTURE_ABSOLUTE_IMPORT" )

        if self._future_print:
            result.append( "CO_FUTURE_PRINT_FUNCTION" )

        if self._barry_bdfl and Utils.python_version >= 300:
            result.append( "CO_FUTURE_BARRY_AS_BDFL" )

        return result



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Assignment related nodes.

All kinds of assignment targets as well as the assignment statement and expression are
located here. These are the core of value control flow.

"""

from .NodeBases import StatementChildrenHavingBase

# Delayed import into multiple branches is not an issue, pylint: disable=W0404

class StatementAssignmentVariable( StatementChildrenHavingBase ):
    kind = "STATEMENT_ASSIGNMENT_VARIABLE"

    named_children = ( "source", "variable_ref" )

    def __init__( self, variable_ref, source, source_ref ):
        assert variable_ref is not None, source_ref
        assert source is not None, source_ref

        assert variable_ref.isTargetVariableRef()

        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "source"       : source,
                "variable_ref" : variable_ref
            },
            source_ref = source_ref
        )

    getTargetVariableRef = StatementChildrenHavingBase.childGetter( "variable_ref" )
    getAssignSource = StatementChildrenHavingBase.childGetter( "source" )

    def mayRaiseException( self, exception_type ):
        return self.getAssignSource().mayRaiseException( exception_type )

    def computeStatement( self, constraint_collection ):
        # Assignment source may re-compute here:
        constraint_collection.onExpression( self.getAssignSource() )

        source = self.getAssignSource()

        constraint_collection.onVariableSet(
            target_node  = self.getTargetVariableRef(),
            value_friend = source.getValueFriend( constraint_collection )
        )

        # TODO: Remove this, it's old.
        return constraint_collection._onStatementAssignmentVariable( self )


class StatementAssignmentAttribute( StatementChildrenHavingBase ):
    kind = "STATEMENT_ASSIGNMENT_ATTRIBUTE"

    named_children = ( "source", "expression" )

    def __init__( self, expression, attribute_name, source, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "expression" : expression,
                "source"     : source,
            },
            source_ref = source_ref
        )

        self.attribute_name = attribute_name

    def getDetails( self ):
        return { "attribute" : self.attribute_name }

    def getDetail( self ):
        return "to attribute %s" % self.attribute_name

    def getAttributeName( self ):
        return self.attribute_name

    def setAttributeName( self, attribute_name ):
        self.attribute_name = attribute_name

    getLookupSource = StatementChildrenHavingBase.childGetter( "expression" )
    getAssignSource = StatementChildrenHavingBase.childGetter( "source" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getAssignSource() )
        source = self.getAssignSource()

        # No assignment will occur, if the assignment source raises, so strip it away.
        if source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = source,
                node       = self
            )

            return result, "new_raise", "Attribute assignment raises exception in assigned value, removed assignment"

        constraint_collection.onExpression( self.getLookupSource() )
        lookup_source = self.getLookupSource()

        if lookup_source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    source,
                    lookup_source
                )
            )

            return result, "new_raise", "Attribute assignment raises exception in attribute source, removed assignment"

        return self, None, None


class StatementAssignmentSubscript( StatementChildrenHavingBase ):
    kind = "STATEMENT_ASSIGNMENT_SUBSCRIPT"

    named_children = ( "source", "expression", "subscript" )

    def __init__( self, expression, subscript, source, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "source"     : source,
                "expression" : expression,
                "subscript"  : subscript
            },
            source_ref = source_ref
        )

    getSubscribed = StatementChildrenHavingBase.childGetter( "expression" )
    getSubscript = StatementChildrenHavingBase.childGetter( "subscript" )
    getAssignSource = StatementChildrenHavingBase.childGetter( "source" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getAssignSource() )
        source = self.getAssignSource()

        # No assignment will occur, if the assignment source raises, so strip it away.
        if source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = source,
                node       = self
            )

            return result, "new_raise", "Subscript assignment raises exception in assigned value, removed assignment"

        constraint_collection.onExpression( self.getSubscribed() )
        subscribed = self.getSubscribed()

        if subscribed.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    source,
                    subscribed
                )
            )

            return result, "new_raise", "Subscript assignment raises exception in subscribed value, removed assignment"

        constraint_collection.onExpression( self.getSubscript() )
        subscript = self.getSubscript()

        if subscript.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    source,
                    subscribed,
                    subscript
                )
            )

            return result, "new_raise", "Subscript assignment raises exception in subscript value, removed assignment"

        return self, None, None


class StatementAssignmentSlice( StatementChildrenHavingBase ):
    kind = "STATEMENT_ASSIGNMENT_SLICE"

    named_children = ( "source", "expression", "lower", "upper" )

    def __init__( self, expression, lower, upper, source, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "source"     : source,
                "expression" : expression,
                "lower"      : lower,
                "upper"      : upper
            },
            source_ref = source_ref
        )

    getLookupSource = StatementChildrenHavingBase.childGetter( "expression" )
    getLower = StatementChildrenHavingBase.childGetter( "lower" )
    getUpper = StatementChildrenHavingBase.childGetter( "upper" )
    getAssignSource = StatementChildrenHavingBase.childGetter( "source" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getAssignSource() )
        source = self.getAssignSource()

        # No assignment will occur, if the assignment source raises, so strip it away.
        if source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = source,
                node       = self
            )

            return result, "new_raise", "Slice assignment raises exception in assigned value, removed assignment"

        constraint_collection.onExpression( self.getLookupSource() )
        lookup_source = self.getLookupSource()

        if lookup_source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    source,
                    lookup_source
                )
            )

            return result, "new_raise", "Slice assignment raises exception in sliced value, removed assignment"

        constraint_collection.onExpression( self.getLower(), allow_none = True )
        lower = self.getLower()

        if lower is not None and lower.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    source,
                    lookup_source,
                    lower
                )
            )

            return result, "new_raise", """\
Slice assignment raises exception in lower slice boundary value, removed assignment."""

        constraint_collection.onExpression( self.getUpper(), allow_none = True )
        upper = self.getUpper()

        if upper is not None and upper.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    source,
                    lookup_source,
                    lower,
                    upper
                )
            )

            return result, "new_raise", """\
Slice assignment raises exception in upper slice boundary value, removed assignment."""

        return self, None, None


class StatementDelVariable( StatementChildrenHavingBase ):
    kind = "STATEMENT_DEL_VARIABLE"

    named_children = ( "variable_ref", )

    def __init__( self, variable_ref, tolerant, source_ref ):
        assert variable_ref is not None
        assert not variable_ref.isExpressionVariableRef()
        assert tolerant is True or tolerant is False

        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "variable_ref" : variable_ref
            },
            source_ref = source_ref
        )

        self.tolerant = tolerant

    def getDetail( self ):
        variable_ref = self.getTargetVariableRef()
        variable = variable_ref.getVariable()

        if variable is not None:
            return "to variable %s" % variable
        else:
            return "to variable %s" % self.getTargetVariableRef()

    def getDetails( self ):
        return { "tolerant" : self.tolerant }

    # TODO: Value propagation needs to make a difference based on this.
    def isTolerant( self ):
        return self.tolerant

    getTargetVariableRef = StatementChildrenHavingBase.childGetter( "variable_ref" )
    def computeStatement( self, constraint_collection ):
        variable = self.getTargetVariableRef().getVariable()

        trace = constraint_collection.getVariableCurrentTrace( variable )

        if trace.isUninitTrace():
            if self.isTolerant():
                return (
                    None,
                    "new_statements",
                    "Removed tolerate del without effect."
                )


        constraint_collection.onVariableDel(
            del_node = self
        )

        return self, None, None

        constraint_collection.onVariableDel(
            del_node = self,
        )

    def mayHaveSideEffects( self ):
        return True

    def mayRaiseException( self, exception_type ):
        if self.tolerant:
            return False
        else:
            return True


class StatementDelAttribute( StatementChildrenHavingBase ):
    kind = "STATEMENT_DEL_ATTRIBUTE"

    named_children = ( "expression", )

    def __init__( self, expression, attribute_name, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "expression" : expression
            },
            source_ref = source_ref
        )

        self.attribute_name = attribute_name

    def getDetails( self ):
        return { "attribute" : self.attribute_name }

    def getDetail( self ):
        return "to attribute %s" % self.attribute_name

    def getAttributeName( self ):
        return self.attribute_name

    def setAttributeName( self, attribute_name ):
        self.attribute_name = attribute_name

    getLookupSource = StatementChildrenHavingBase.childGetter( "expression" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getLookupSource() )
        lookup_source = self.getLookupSource()

        if lookup_source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            return makeStatementExpressionOnlyReplacementNode(
                expression = lookup_source,
                node       = self
            )

        return self, None, None


class StatementDelSubscript( StatementChildrenHavingBase ):
    kind = "STATEMENT_DEL_SUBSCRIPT"

    named_children = ( "expression", "subscript" )

    def __init__( self, expression, subscript, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "expression" : expression,
                "subscript"  : subscript
            },
            source_ref = source_ref
        )

    getSubscribed = StatementChildrenHavingBase.childGetter( "expression" )
    getSubscript = StatementChildrenHavingBase.childGetter( "subscript" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getSubscribed() )
        subscribed = self.getSubscribed()

        if subscribed.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = subscribed,
                node       = self
            )

            return result, "new_raise", "Subscript del raises exception in subscribed value, removed del"


        constraint_collection.onExpression( self.getSubscript() )
        subscript = self.getSubscript()

        if subscript.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    subscribed,
                    subscript
                )
            )

            return result, "new_raise", "Subscript del raises exception in subscribt value, removed del"

        return self, None, None


class StatementDelSlice( StatementChildrenHavingBase ):
    kind = "STATEMENT_DEL_SLICE"

    named_children = ( "expression", "lower", "upper" )

    def __init__( self, expression, lower, upper, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "expression" : expression,
                "lower"      : lower,
                "upper"      : upper
            },
            source_ref = source_ref
        )

    getLookupSource = StatementChildrenHavingBase.childGetter( "expression" )
    getLower = StatementChildrenHavingBase.childGetter( "lower" )
    getUpper = StatementChildrenHavingBase.childGetter( "upper" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getLookupSource() )
        lookup_source = self.getLookupSource()

        if lookup_source.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = lookup_source,
                node       = self
            )

            return result, "new_raise", "Slice del raises exception in sliced value, removed del"


        constraint_collection.onExpression( self.getLower(), allow_none = True )
        lower = self.getLower()

        if lower is not None and lower.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    lookup_source,
                    lower
                )
            )

            return result, "new_raise", "Slice del raises exception in lower slice boundary value, removed del"

        constraint_collection.onExpression( self.getUpper(), allow_none = True )
        upper = self.getUpper()

        if upper is not None and upper.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = (
                    lookup_source,
                    lower,
                    upper
                )
            )

            return result, "new_raise", "Slice del raises exception in upper slice boundary value, removed del"

        return self, None, None



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Loop nodes.

There are for and loop nodes, and the break/continue statements for it. Loops are a very
difficult topic, and it might be that the two forms need to be reduced to one common form,
which is more general, the 'Forever' loop, with breaks

"""

from .NodeBases import (
    StatementChildrenHavingBase,
    NodeBase
)


class StatementLoop( StatementChildrenHavingBase ):
    kind = "STATEMENT_LOOP"

    named_children = ( "frame", )

    def __init__( self, body, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "frame" : body
            },
            source_ref = source_ref
        )

        self.break_exception = False
        self.continue_exception = False

    getLoopBody = StatementChildrenHavingBase.childGetter( "frame" )

    def markAsExceptionContinue( self ):
        self.continue_exception = True

    def markAsExceptionBreak( self ):
        self.break_exception = True

    def needsExceptionContinue( self ):
        return self.continue_exception

    def needsExceptionBreak( self ):
        return self.break_exception

    def computeStatement( self, constraint_collection ):
        loop_body = self.getLoopBody()

        if loop_body is not None:
            from nuitka.optimizations.ConstraintCollections import ConstraintCollectionLoop

            loop_collection = ConstraintCollectionLoop( constraint_collection )
            loop_collection.process( loop_body )

            # Might be changed.
            loop_body = self.getLoopBody()

        # Consider trailing "continue" statements, these have no effect, so we can remove
        # them.
        if loop_body is not None:
            assert loop_body.isStatementsSequence()

            statements = loop_body.getStatements()
            assert statements # Cannot be empty

            last_statement = statements[-1]
            if last_statement.isStatementContinueLoop():
                loop_body.removeStatement( last_statement )
                statements = loop_body.getStatements()

                if not statements:
                    loop_body.replaceWith( None )
                    loop_body = None

                constraint_collection.signalChange(
                    "new_statements",
                    last_statement.getSourceReference(),
                    "Removed continue as last statement of loop."
                )

        # Consider leading "break" statements, they should be the only, and should lead to
        # removing the whole loop statement. Trailing "break" statements could also be
        # handled, but that would need to consider if there are other "break" statements
        # too. Numbering loop exits is nothing we have yet.
        if loop_body is not None:
            assert loop_body.isStatementsSequence()

            statements = loop_body.getStatements()
            assert statements # Cannot be empty

            if len( statements ) == 1 and statements[-1].isStatementBreakLoop():
                return None, "new_statements", "Removed loop immediately broken."

        return self, None, None


class StatementContinueLoop( NodeBase ):
    kind = "STATEMENT_CONTINUE_LOOP"

    def __init__( self, source_ref ):
        NodeBase.__init__( self, source_ref = source_ref )

        self.exception_driven = False

    def isStatementAborting( self ):
        return True

    def markAsExceptionDriven( self ):
        self.exception_driven = True

    def isExceptionDriven( self ):
        return self.exception_driven

    def computeStatement( self, constraint_collection ):
        # This statement being aborting, will already tell everything. TODO: The fine
        # difference that this jumps to loop start for sure, should be represented somehow
        # one day.
        return self, None, None


class StatementBreakLoop( NodeBase ):
    kind = "STATEMENT_BREAK_LOOP"

    def __init__( self, source_ref ):
        NodeBase.__init__( self, source_ref = source_ref )

        self.exception_driven = False

    def isStatementAborting( self ):
        return True

    def markAsExceptionDriven( self ):
        self.exception_driven = True

    def isExceptionDriven( self ):
        return self.exception_driven

    def computeStatement( self, constraint_collection ):
        # This statement being aborting, will already tell everything. TODO: The fine
        # difference that this exits the loop for sure, should be represented somehow one
        # day.
        return self, None, None



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Nodes for statements.

"""

from .NodeBases import StatementChildrenHavingBase

from nuitka.Utils import python_version

def mergeStatements( statements ):
    """ Helper function that merges nested statement sequences. """
    merged_statements = []

    for statement in statements:
        if statement.isStatement() or statement.isStatementsFrame():
            merged_statements.append( statement )
        elif statement.isStatementsSequence():
            merged_statements.extend( mergeStatements( statement.getStatements() ) )
        else:
            assert False, statement

    return merged_statements


class StatementsSequence( StatementChildrenHavingBase ):
    kind = "STATEMENTS_SEQUENCE"

    named_children = ( "statements", )

    def __init__( self, statements, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "statements" : tuple( statements )
            },
            source_ref = source_ref
        )

    getStatements = StatementChildrenHavingBase.childGetter( "statements" )
    setStatements = StatementChildrenHavingBase.childSetterNotNone( "statements" )

    def getDetails( self ):
        if self.getStatements():
            return { "statement_count" : len( self.getStatements() ) }
        else:
            return { "statement_count" : 0 }

    def setChild( self, name, value ):
        assert name == "statements"

        assert None not in value, value

        return StatementChildrenHavingBase.setChild(
            self,
            name  = name,
            value = mergeStatements( value )
        )

    # Overloading automatic check, so that derived ones know it too.
    def isStatementsSequence( self ):
        # Virtual method, pylint: disable=R0201,W0613

        return True

    def trimStatements( self, statement ):
        assert statement.parent is self

        old_statements = list( self.getStatements() )
        assert statement in old_statements, ( statement, self )

        new_statements = old_statements[ : old_statements.index( statement ) + 1 ]

        self.setChild( "statements", new_statements )

    def removeStatement( self, statement ):
        assert statement.parent is self

        statements = list( self.getStatements() )
        statements.remove( statement )

        self.setChild( "statements", statements )

    def mergeStatementsSequence( self, statement_sequence ):
        assert statement_sequence.parent is self

        old_statements = list( self.getStatements() )
        assert statement_sequence in old_statements, ( statement_sequence, self )

        merge_index =  old_statements.index( statement_sequence )

        new_statements = tuple( old_statements[ : merge_index ] )   + \
                         statement_sequence.getStatements()         + \
                         tuple( old_statements[ merge_index + 1 : ] )

        self.setChild( "statements", new_statements )


    def mayHaveSideEffects( self ):
        assert constraint_collection is None

        # Statement sequences have a side effect if one of the statements does.
        for statement in self.getStatements():
            if statement.mayHaveSideEffects():
                return True
        else:
            return False

    def isStatementAborting( self ):
        return self.getStatements()[-1].isStatementAborting()


class StatementsFrame( StatementsSequence ):
    kind = "STATEMENTS_FRAME"

    def __init__( self, statements, guard_mode, code_name, var_names, arg_count,
                  kw_only_count, has_starlist, has_stardict, source_ref ):
        StatementsSequence.__init__(
            self,
            statements = statements,
            source_ref = source_ref
        )

        self.var_names = tuple( var_names )
        self.code_name = code_name

        self.kw_only_count = kw_only_count
        self.arg_count = arg_count

        self.guard_mode = guard_mode

        self.has_starlist = has_starlist
        self.has_stardict = has_stardict

    def getDetails( self ):
        result = {
            "code_name"  : self.code_name,
            "var_names"  : ", ".join( self.var_names ),
            "guard_mode" : self.guard_mode
        }

        if python_version >= 300:
            result[ "kw_only_count" ] = self.kw_only_count

        return result

    def needsLineNumber( self ):
        return False

    def getGuardMode( self ):
        return self.guard_mode

    def getVarNames( self ):
        return self.var_names

    def getCodeObjectName( self ):
        return self.code_name

    def getKwOnlyParameterCount( self ):
        return self.kw_only_count

    def getArgumentCount( self ):
        return self.arg_count

    def makeCloneAt( self, source_ref ):
        assert False

    def getCodeObjectHandle( self, context ):
        provider = self.getParentVariableProvider()

        # TODO: Why do this accessing a node, do this outside.
        from nuitka.codegen.CodeObjectCodes import getCodeObjectHandle

        return getCodeObjectHandle(
            context       = context,
            filename      = self.source_ref.getFilename(),
            var_names     = self.getVarNames(),
            arg_count     = self.getArgumentCount(),
            kw_only_count = self.getKwOnlyParameterCount(),
            line_number   = 0
                              if provider.isPythonModule() else
                            self.source_ref.getLineNumber(),
            code_name     = self.getCodeObjectName(),
            is_generator  = provider.isExpressionFunctionBody() and \
                            provider.isGenerator(),
            is_optimized  = not provider.isPythonModule() and \
                            not provider.isClassDictCreation() and \
                            not context.hasLocalsDict(),
            has_starlist  = self.has_starlist,
            has_stardict  = self.has_stardict
        )


class StatementExpressionOnly( StatementChildrenHavingBase ):
    kind = "STATEMENT_EXPRESSION_ONLY"

    named_children = ( "expression", )

    def __init__( self, expression, source_ref ):
        assert expression.isExpression()

        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "expression" : expression
            },
            source_ref = source_ref
        )

    def getDetail( self ):
        return "expression %s" % self.getExpression()

    def mayHaveSideEffects( self ):
        return self.getExpression().mayHaveSideEffects()

    getExpression = StatementChildrenHavingBase.childGetter( "expression" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getExpression() )
        expression = self.getExpression()

        # Side effects can  become statements.
        if expression.isExpressionSideEffects():
            from .NodeMakingHelpers import makeStatementOnlyNodesFromExpressions

            result = makeStatementOnlyNodesFromExpressions(
                expressions = expression.getSideEffects() + \
                              ( expression.getExpression(), )
            )

            return result, "new_statements", """\
Turned side effects of expression only statement into statements."""

        elif not expression.mayHaveSideEffects():
            return None, "new_statements", "Removed statement without effect."

        elif expression.isExpressionRaiseException():
            from .ExceptionNodes import StatementRaiseExceptionImplicit

            result = StatementRaiseExceptionImplicit(
                exception_type  = expression.getExceptionType(),
                exception_value = expression.getExceptionValue(),
                exception_trace = None,
                exception_cause = None,
                source_ref      = expression.getSourceReference()
            )

            return result, "new_raise", """\
Propgated implict raise expression to raise statement."""
        elif python_version < 300 and expression.isExpressionBuiltinExecfile():
            # In this case, the copy-back must be done and will only be done
            # correctly by the code for exec statements.
            provider = self.getParentVariableProvider()

            if provider.isExpressionFunctionBody() and provider.isClassDictCreation():
                from .ExecEvalNodes import StatementExec

                result = StatementExec(
                    source_code = expression.getSourceCode(),
                    globals_arg = expression.getGlobals(),
                    locals_arg  = expression.getLocals(),
                    source_ref  = expression.getSourceReference()
                )

                return result, "new_statements", """\
Changed execfile to exec on class level"""
        elif python_version >= 300 and expression.isExpressionBuiltinExec():
            if self.getParentVariableProvider().isEarlyClosure():
                from .ExecEvalNodes import StatementExec

                result = StatementExec(
                    source_code = expression.getSourceCode(),
                    globals_arg = expression.getGlobals(),
                    locals_arg  = expression.getLocals(),
                    source_ref  = expression.getSourceReference()
                )

                return result, "new_statements", """\
Replaced builtin exec call to exec statement in early closure context."""

        return self, None, None



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Low level constant code generation.

"""

from .Pickling import getStreamedConstant
from .CppStrings import encodeString

from .TupleCodes import addMakeTupleUse
from .ListCodes import addMakeListUse
from .DictCodes import addMakeDictUse

# pylint: disable=W0622
from ..__past__ import unicode, long, iterItems
# pylint: enable=W0622

from ..Constants import HashableConstant, constant_builtin_types

import re, struct

def getConstantHandle( context, constant ):
    return context.getConstantHandle( constant )

def getConstantCode( context, constant ):
    constant_identifier = context.getConstantHandle(
        constant = constant
    )

    return constant_identifier.getCode()

# TODO: The determination of this should already happen in Building or in a
# helper not during code generation.
_match_attribute_names = re.compile( r"[a-zA-Z_][a-zA-Z0-9_]*$" )

def getConstantCodeName( context, constant ):
    return context.getConstantHandle( constant, real_use = False ).getCode()

def _isAttributeName( value ):
    return _match_attribute_names.match( value )

_needs_pickle = False

def needsPickleInit():
    return _needs_pickle

def _getUnstreamCode( constant_value, constant_identifier ):
    saved = getStreamedConstant(
        constant_value = constant_value
    )

    assert type( saved ) is bytes

    # We need to remember having to use pickle, pylint: disable=W0603
    global _needs_pickle
    _needs_pickle = True

    return "%s = UNSTREAM_CONSTANT( %s, %d );" % (
        constant_identifier,
        encodeString( saved ),
        len( saved )
    )


def _packFloat( value ):
    return struct.pack( "<d", value )

def _addConstantInitCode( context, emit, constant_type, constant_value,
                          constant_identifier ):
    # This has many cases, that all return, and do a lot.
    #  pylint: disable=R0911,R0912,R0915

    # Use shortest code for ints and longs, except when they are big, then fall
    # fallback to pickling. TODO: Avoid the use of pickle even for larger
    # values.
    if constant_type is int and abs( constant_value ) < 2**31:
        emit(
            "%s = PyInt_FromLong( %s );" % (
                constant_identifier,
                constant_value
            )
        )

        return

    # See above, same for long values. Note: These are of course not existant
    # with Python3 which would have covered it before.
    if constant_type is long and abs( constant_value ) < 2**31:
        emit (
            "%s = PyLong_FromLong( %s );" % (
                constant_identifier,
                constant_value
            )
        )

        return

    # Strings that can be encoded as UTF-8 are done more or less directly. When
    # they cannot be expressed as UTF-8, that is rare not we can indeed use
    # pickling.
    if constant_type is str:
        if str is not unicode:
            emit(
                "%s = UNSTREAM_STRING( %s, %d, %d );assert( %s );" % (
                    constant_identifier,
                    encodeString( constant_value ),
                    len( constant_value ),
                    1 if _isAttributeName( constant_value ) else 0,
                    constant_identifier
                )
            )

            return
        else:
            try:
                encoded = constant_value.encode( "utf-8" )

                emit(
                    "%s = UNSTREAM_STRING( %s, %d, %d );assert( %s );" % (
                        constant_identifier,
                        encodeString( encoded ),
                        len( encoded ),
                        1 if _isAttributeName( constant_value ) else 0,
                        constant_identifier
                    )
                )

                return
            except UnicodeEncodeError:
                # So fall back to below code, which will unstream it then.
                pass

    if constant_type is float:
        emit(
            "%s = UNSTREAM_FLOAT( %s );" % (
                constant_identifier,
                encodeString( _packFloat( constant_value ) )
            )
        )

        return

    if constant_value is None:
        return

    if constant_value is False:
        return

    if constant_value is True:
        return

    if constant_type is dict:
        if constant_value == {}:
            emit( "%s = PyDict_New();" % constant_identifier )
        else:
            length = len( constant_value )
            addMakeDictUse( length )

            for key, value in iterItems( constant_value ):
                _addConstantInitCode(
                    emit                = emit,
                    constant_type       = type( key ),
                    constant_value      = key,
                    constant_identifier = getConstantCodeName( context, key ),
                    context             = context
                )
                _addConstantInitCode(
                    emit                = emit,
                    constant_type       = type( value ),
                    constant_value      = value,
                    constant_identifier = getConstantCodeName( context, value ),
                    context             = context
                )

            emit(
                "%s = MAKE_DICT%d( %s );" % (
                    constant_identifier,
                    length,
                    ", ".join(
                        "%s, %s" % (
                            getConstantCodeName( context, value ),
                            getConstantCodeName( context, key )
                        )
                        for key, value
                        in
                        iterItems( constant_value )
                    )
                )
            )

        return

    if constant_type is tuple:
        if constant_value == ():
            emit( "%s = PyTuple_New( 0 );" % constant_identifier )
        else:
            length = len( constant_value )
            addMakeTupleUse( length )

            # Make elements earlier than tuple itself.
            for element in constant_value:
                _addConstantInitCode(
                    emit                = emit,
                    constant_type       = type( element ),
                    constant_value      = element,
                    constant_identifier = getConstantCodeName( context, element ),
                    context             = context
                )

            emit(
                "%s = MAKE_TUPLE%d( %s );" % (
                    constant_identifier,
                    length,
                    ", ".join(
                        getConstantCodeName( context, element )
                        for element
                        in
                        constant_value
                    )
                )
            )

        return

    if constant_type is list:
        if constant_value == []:
            emit( "%s = PyList_New( 0 );" % constant_identifier )
        else:
            length = len( constant_value )
            addMakeListUse( length )

            # Make elements earlier than list itself.
            for element in constant_value:
                _addConstantInitCode(
                    emit                = emit,
                    constant_type       = type( element ),
                    constant_value      = element,
                    constant_identifier = getConstantCodeName( context, element ),
                    context             = context
                )

            emit(
                "%s = MAKE_LIST%d( %s );" % (
                    constant_identifier,
                    length,
                    ", ".join(
                        getConstantCodeName( context, element )
                        for element
                        in
                        constant_value
                    )
                )
            )

        return

    if constant_type is list and constant_value == []:
        emit( "%s = PyList_New( 0 );" % constant_identifier )

        return

    if constant_type is set and constant_value == set():
        emit( "%s = PySet_New( NULL );" % constant_identifier )

        return

    if constant_type in ( set, frozenset, complex, unicode, int, long, bytes, range ):
        emit(  _getUnstreamCode( constant_value, constant_identifier ) )

        return

    if constant_value in constant_builtin_types:
        return

    assert False, (type(constant_value), constant_value, constant_identifier)

def _lengthKey( value ):
    return len( value[1] ), value[1]

the_contained_constants = {}

def getConstantsInitCode( context ):
    # There are many cases for constants to be created in the most efficient
    # way, pylint: disable=R0912

    statements = []

    all_constants = the_contained_constants
    all_constants.update( context.getConstants() )

    def receiveStatement( statement ):
        assert statement is not None

        if statement not in statements:
            statements.append( statement )

    for ( constant_type, constant_value ), constant_identifier in \
          sorted( all_constants.items(), key = _lengthKey ):
        _addConstantInitCode(
            emit                = receiveStatement,
            constant_type       = constant_type,
            constant_value      = constant_value.getConstant(),
            constant_identifier = constant_identifier,
            context             = context
        )

    return statements

def getConstantsDeclCode( context, for_header ):
    # There are many cases for constants of different types.
    # pylint: disable=R0912
    statements = []

    constants = context.getConstants()

    contained_constants = {}

    def considerForDeferral( constant_value ):
        if constant_value is None:
            return

        if constant_value is False:
            return

        if constant_value is True:
            return

        constant_type = type( constant_value )

        if constant_type is type:
            return

        key = constant_type, HashableConstant( constant_value )

        if key not in contained_constants:
            constant_identifier = getConstantCodeName( context, constant_value )

            contained_constants[ key ] = constant_identifier

            if constant_type in ( tuple, list ):
                for element in constant_value:
                    considerForDeferral( element )

            if constant_type is dict:
                for key, value in iterItems( constant_value ):
                    considerForDeferral( key )
                    considerForDeferral( value )


    for ( constant_type, constant_value ), constant_identifier in sorted( constants.items(), key = _lengthKey ):
        # Need not declare built-in types.
        if constant_type is type:
            continue

        declaration = "PyObject *%s;" % constant_identifier

        if for_header:
            declaration = "extern " + declaration
        else:
            if constant_type in ( tuple, dict, list ):
                considerForDeferral( constant_value.getConstant() )

        statements.append( declaration )

    for key, value in sorted( contained_constants.items(), key = _lengthKey ):
        if key not in constants:
            declaration = "static PyObject *%s;" % value

            statements.append( declaration )

    if not for_header:
        global the_contained_constants
        the_contained_constants = contained_constants

    return statements



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Builtin iterator nodes.

These play a role in for loops, and in unpacking. They can something be predicted to
succeed or fail, in which case, code can become less complex. The length of things is an
important optimization issue.
"""

from .NodeBases import (
    ExpressionBuiltinSingleArgBase,
    ExpressionChildrenHavingBase,
    StatementChildrenHavingBase
)

from .ValueFriends import ValueFriendBase

from .SideEffectNodes import ExpressionSideEffects

from nuitka.optimizations import BuiltinOptimization

from nuitka import Options


class ExpressionBuiltinLen( ExpressionBuiltinSingleArgBase ):
    kind = "EXPRESSION_BUILTIN_LEN"

    builtin_spec = BuiltinOptimization.builtin_len_spec

    def getIntegerValue( self ):
        return self.getValue().getIterationLength()

    def computeExpression( self, constraint_collection ):
        from .NodeMakingHelpers import makeConstantReplacementNode, wrapExpressionWithNodeSideEffects

        new_node, change_tags, change_desc = ExpressionBuiltinSingleArgBase.computeExpression(
            self,
            constraint_collection = constraint_collection
        )

        if new_node is self:
            arg_length = self.getIntegerValue()

            if arg_length is not None:
                change_tags = "new_constant"
                change_desc = "Predicted len argument"

                new_node = wrapExpressionWithNodeSideEffects(
                    new_node = makeConstantReplacementNode( arg_length, self ),
                    old_node = self.getValue()
                )

                if new_node.isExpressionSideEffects(): # false alarm pylint: disable=E1101
                    change_desc += " maintaining side effects"

        return new_node, change_tags, change_desc


class ValueFriendBuiltinIter1( ValueFriendBase ):
    def __init__( self, iterated ):
        ValueFriendBase.__init__( self )

        self.iterated = iterated
        self.iter_length = None
        self.consumed = 0

        self.used = False

    def __eq__( self, other ):
        if self.__class__ is not other.__class__:
            return False

        return self.iterated == other.iterated and self.consumed == other.consumed

    def mayProvideReference( self ):
        # Virtual method overload, where it's fixed by type, pylint: disable=R0201
        return True

    def isKnownToBeIterableAtMin( self, count ):
        if self.iter_length is None:
            self.iter_length = self.iterated.getIterationLength()

        return self.iter_length is not None and self.iter_length - self.consumed >= count

    def isKnownToBeIterableAtMax( self, count ):
        if self.iter_length is None:
            self.iter_length = self.iterated.getIterationLength()

        return self.iter_length is not None and self.iter_length - self.consumed <= count

    def getIterationNext( self ):
        # print self.iterated, self.consumed, self.iterated.getVisitableNodes()

        if self.iterated.canPredictIterationValues():
            result = self.iterated.getIterationValue( self.consumed )
        else:
            result = None

        self.consumed += 1

        return result

    def markAsUsed( self ):
        self.used = True

    def onRelease( self, constraint_collection ):
        # print "onRelease", self
        pass


class ExpressionBuiltinIter1( ExpressionBuiltinSingleArgBase ):
    kind = "EXPRESSION_BUILTIN_ITER1"

    def getValueFriend( self, constraint_collection ):
        return ValueFriendBuiltinIter1( self.getValue().getValueFriend( constraint_collection ) )

    def computeExpression( self, constraint_collection ):
        value = self.getValue()

        if value.isIteratorMaking():
            return value, "new_builtin", "Eliminated useless iterator creation"
        else:
            return self, None, None

    def isIteratorMaking( self ):
        return True

    def isKnownToBeIterable( self, count ):
        if count is None:
            return True

        # TODO: Should ask value if it is.
        return None

    def getIterationLength( self ):
        return self.getValue().getIterationLength()

    def extractSideEffects( self ):
        # Iterator making is the side effect itself.
        if self.getValue().isCompileTimeConstant():
            return ()
        else:
            return ( self, )


    def mayHaveSideEffects( self ):
        if self.getValue().isCompileTimeConstant():
            return self.getValue().isKnownToBeIterable( None )

        return True


class ExpressionBuiltinNext1( ExpressionBuiltinSingleArgBase ):
    kind = "EXPRESSION_BUILTIN_NEXT1"

    def getDetails( self ):
        return {
            "iter" : self.getValue()
        }

    def makeCloneAt( self, source_ref ):
        return self.__class__(
            value      = self.getValue(),
            source_ref = source_ref
        )

    def computeExpression( self, constraint_collection ):
        # TODO: Enable below code once safer.
        if True or not Options.isExperimental():
            return self, None, None

        target = self.getValue().getValueFriend( constraint_collection )

        if target.isKnownToBeIterableAtMin( 1 ):
            value = target.getIterationNext()

            if value is not None:
                if value.isNode() and not self.parent.isStatementExpressionOnly():
                    # As a side effect, keep the iteration, later checks may depend on it,
                    # or if absent, optimizations will remove it.
                    if not self.parent.isExpressionSideEffects():
                        value = ExpressionSideEffects(
                            expression   = value.makeCloneAt(
                                source_ref = self.getSourceReference()
                            ),
                            side_effects = (
                                self.makeCloneAt(
                                    source_ref = self.getSourceReference()
                                ),
                            ),
                            source_ref   = self.getSourceReference()
                        )

                    return value, "new_expression", "Predicted next iteration result"
            else:
                assert False, target

        return self, None, None


class ExpressionSpecialUnpack( ExpressionBuiltinNext1 ):
    kind = "EXPRESSION_SPECIAL_UNPACK"

    def __init__( self, value, count, source_ref ):
        ExpressionBuiltinNext1.__init__(
            self,
            value      = value,
            source_ref = source_ref
        )

        self.count = count

    def makeCloneAt( self, source_ref ):
        return self.__class__(
            value      = self.getValue(),
            count      = self.getCount(),
            source_ref = source_ref
        )

    def getDetails( self ):
        result = ExpressionBuiltinNext1.getDetails( self )
        result[ "element_index" ] = self.getCount()

        return result

    def getCount( self ):
        return self.count


class StatementSpecialUnpackCheck( StatementChildrenHavingBase ):
    kind = "STATEMENT_SPECIAL_UNPACK_CHECK"

    named_children = ( "iterator", )

    def __init__( self, iterator, count, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "iterator" : iterator
            },
            source_ref = source_ref
        )

        self.count = count

    def getDetails( self ):
        return {
            "count" : self.getCount(),
        }

    def getCount( self ):
        return self.count

    getIterator = StatementChildrenHavingBase.childGetter( "iterator" )

    def computeStatement( self, constraint_collection ):
        constraint_collection.onExpression( self.getIterator() )
        iterator = self.getIterator()

        if iterator.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = iterator,
                node       = self
            )

            return result, "new_raise", "Explicit raise already raises implicitely building exception type"

        # Remove the check if it can be decided at compile time.
        if iterator.isKnownToBeIterableAtMax( 0 ):
            return None, "new_statements", "Determined iteration end check to be always true."

        return self, None, None


class ExpressionBuiltinIter2( ExpressionChildrenHavingBase ):
    kind = "EXPRESSION_BUILTIN_ITER2"

    named_children = ( "callable", "sentinel", )

    # Need to accept 'callable' keyword argument, that is just the API of iter,
    # pylint: disable=W0622

    def __init__( self, callable, sentinel, source_ref ):
        ExpressionChildrenHavingBase.__init__(
            self,
            values = {
                "callable" : callable,
                "sentinel" : sentinel,
            },
            source_ref = source_ref
        )

    getCallable = ExpressionChildrenHavingBase.childGetter( "callable" )
    getSentinel = ExpressionChildrenHavingBase.childGetter( "sentinel" )

    def computeExpression( self, constraint_collection ):
        # TODO: The "callable" should be investigated here, pylint: disable=W0613

        return self, None, None

    def isIteratorMaking( self ):
        return True


class ExpressionBuiltinNext2( ExpressionChildrenHavingBase ):
    kind = "EXPRESSION_BUILTIN_NEXT2"

    named_children = ( "iterator", "default", )

    def __init__( self, iterator, default, source_ref ):
        ExpressionChildrenHavingBase.__init__(
            self,
            values = {
                "iterator" : iterator,
                "default"  : default,
            },
            source_ref = source_ref
        )

    getIterator = ExpressionChildrenHavingBase.childGetter( "iterator" )
    getDefault = ExpressionChildrenHavingBase.childGetter( "default" )

    def computeExpression( self, constraint_collection ):
        # TODO: The "iterator" should be investigated here, pylint: disable=W0613

        return self, None, None



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Low level variable code generation.

"""

from nuitka import Variables

from .Identifiers import (
    MaybeModuleVariableIdentifier,
    ClosureVariableIdentifier,
    ModuleVariableIdentifier,
    LocalVariableIdentifier,
    TempVariableIdentifier,
    TempObjectIdentifier,
    encodeNonAscii
)

from .ConstantCodes import getConstantCode

def getVariableHandle( context, variable ):
    assert isinstance( variable, Variables.Variable ), variable

    var_name = variable.getName()

    if variable.isLocalVariable() or variable.isClassVariable():
        return LocalVariableIdentifier(
            var_name     = var_name,
            from_context = context.getFunction().isGenerator()
        )
    elif variable.isTempVariableReference():
        if not variable.getOwner().isStatementTempBlock():
            variable = variable.getReferenced()

        if not variable.getReferenced().getNeedsFree():
            return TempObjectIdentifier(
                var_name = var_name
            )
        else:
            return TempVariableIdentifier(
                var_name = var_name
            )
    elif variable.isClosureReference():
        function = context.getFunction()

        # Context access is variant depending on if that's a created function or
        # not. For generators, they even share closure variables in the common
        # context.
        if function.needsCreation():
            if function.isGenerator():
                from_context = "_python_context->common_context->"
            else:
                from_context = "_python_context->"
        else:
            if function.isGenerator():
                from_context = "_python_context->"
            else:
                from_context = ""

        return ClosureVariableIdentifier(
            var_name     = var_name,
            from_context = from_context
        )
    elif variable.isMaybeLocalVariable():
        context.addGlobalVariableNameUsage( var_name )

        assert context.getFunction().hasLocalsDict(), context

        return MaybeModuleVariableIdentifier(
            var_name         = var_name,
            module_code_name = context.getModuleCodeName()
        )
    elif variable.isModuleVariable():
        context.addGlobalVariableNameUsage(
            var_name = var_name
        )

        return ModuleVariableIdentifier(
            var_name         = var_name,
            module_code_name = context.getModuleCodeName()
        )
    else:
        assert False, variable

def getVariableCode( context, variable ):
    var_identifier = getVariableHandle(
        context  = context,
        variable = variable
    )

    return var_identifier.getCode()

def getLocalVariableInitCode( context, variable, init_from = None,
                              in_context = False ):
    # This has many cases to deal with, so there need to be a lot of branches.
    # pylint: disable=R0912

    assert not variable.isModuleVariable()

    assert init_from is None or hasattr( init_from, "getCodeTemporaryRef" )

    result = variable.getDeclarationTypeCode( in_context )

    # For pointer types, we don't have to separate with spaces.
    if not result.endswith( "*" ):
        result += " "

    store_name = variable.getMangledName()

    if not in_context:
        result += "_"

    result += encodeNonAscii( variable.getCodeName() )

    if not in_context:
        if variable.isTempVariable():
            if init_from is None:
                result += " = " + variable.getDeclarationInitValueCode()
            elif not variable.getNeedsFree():
                result += " = %s" % init_from.getCodeTemporaryRef()
            else:
                result += "( %s )" % init_from.getCodeExportRef()
        else:
            result += "( "

            result += "%s" % getConstantCode(
                context  = context,
                constant = store_name
            )

            if init_from is not None:
                if context.hasLocalsDict():
                    if init_from.getCheapRefCount() == 0:
                        result += ", %s" % init_from.getCodeTemporaryRef()
                    else:
                        result += ", %s" % init_from.getCodeExportRef()

                        if not variable.isParameterVariable():
                            result += ", true"
                else:
                    result += ", %s" % init_from.getCodeExportRef()

            result += " )"

    result += ";"

    return result



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Recursion into other modules.

"""

from nuitka import Utils, Importing, ModuleRegistry

from . import ImportCache, Building

from logging import info, warning

def recurseTo( module_package, module_filename, module_relpath ):
    if not ImportCache.isImportedModuleByPath( module_relpath ):
        module, source_ref, source_filename = Building.decideModuleTree(
            filename = module_filename,
            package  = module_package,
            is_top   = False,
            is_main  = False
        )

        # Check if the module name is known. In order to avoid duplicates,
        # learn the new filename, and continue build if its not.
        if not ImportCache.isImportedModuleByName( module.getFullName() ):
            info(
                "Recurse to import '%s' from %s",
                module.getFullName(),
                module_relpath
            )

            try:
                Building.createModuleTree(
                    module          = module,
                    source_ref      = source_ref,
                    source_filename = source_filename,
                    is_main         = False
                )
            except ( SyntaxError, IndentationError ) as e:
                if module_filename not in Importing.warned_about:
                    Importing.warned_about.add( module_filename )

                    warning(
                        """\
Cannot recurse to import module '%s' (%s) because of '%s'""",
                        module_relpath,
                        module_filename,
                        e.__class__.__name__
                    )

                return None, False

            ImportCache.addImportedModule(
                module_relpath,
                module
            )

            is_added = True
        else:
            ImportCache.addImportedModule(
                module_relpath,
                ImportCache.getImportedModuleByName( module.getFullName() )
            )

            module = ImportCache.getImportedModuleByName(
                module.getFullName()
            )

            is_added = False

        assert not module_relpath.endswith( "/__init__.py" )

        return module, is_added
    else:
        return ImportCache.getImportedModuleByPath( module_relpath ), False


def considerFilename( module_filename, module_package ):
    assert module_package is None or ( type( module_package ) is str and module_package != "" )

    module_filename = Utils.normpath( module_filename )

    if Utils.isDir( module_filename ):
        module_filename = Utils.abspath( module_filename )

        module_name = Utils.basename( module_filename )
        module_relpath = Utils.relpath( module_filename )

        return module_filename, module_relpath, module_name
    elif module_filename.endswith( ".py" ):
        module_name = Utils.basename( module_filename )[:-3]
        module_relpath = Utils.relpath( module_filename )

        return module_filename, module_relpath, module_name
    else:
        return None

def _checkPluginPath( plugin_filename, module_package ):
    plugin_info = considerFilename(
        module_package  = module_package,
        module_filename = plugin_filename
    )

    if plugin_info is not None:
        module, added = recurseTo(
            module_filename = plugin_info[0],
            module_relpath  = plugin_info[1],
            module_package  = module_package
        )

        if module:
            if not added:
                warning(
                    "Recursed to %s '%s' at '%s' twice.",
                    "package" if module.isPythonPackage() else "module",
                    module.getName(),
                    plugin_info[0]
                )

            ModuleRegistry.addRootModule( module )

            if module.isPythonPackage():
                package_dir = Utils.dirname( module.getFilename() )

                for sub_path, sub_filename in Utils.listDir( package_dir ):
                    if sub_filename == "__init__.py":
                        continue

                    assert sub_path != plugin_filename, package_dir

                    if Importing.isPackageDir( sub_path ) or sub_path.endswith( ".py" ):
                        _checkPluginPath( sub_path, module.getFullName() )


        else:
            warning( "Failed to include module from '%s'.", plugin_info[0] )

def checkPluginPath( plugin_filename, module_package ):
    plugin_info = considerFilename(
        module_package  = module_package,
        module_filename = plugin_filename
    )

    if plugin_info is not None:
        # File or package, handle that.
        if Utils.isFile( plugin_info[0] ) or Importing.isPackageDir( plugin_info[0] ):
            _checkPluginPath( plugin_filename, module_package )
        elif Utils.isDir( plugin_info[0] ):
            for sub_path, sub_filename in Utils.listDir( plugin_info[0] ):
                assert sub_filename != "__init__.py"

                if Importing.isPackageDir( sub_path ) or sub_path.endswith( ".py" ):
                    _checkPluginPath( sub_path, None )
        else:
            warning( "Failed to include module from '%s'.", plugin_info[0] )



#     Copyright 2013, Kay Hayen, mailto:kay.hayen@gmail.com
#
#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#
""" Conditional nodes.

These is the conditional expression '(a if b else c)' and the conditional
statement, 'if a: ... else: ...' and there is no 'elif', because that is
expressed via nesting of conditional statements.
"""

from .NodeBases import ExpressionChildrenHavingBase, StatementChildrenHavingBase

# Delayed import into multiple branches is not an issue, pylint: disable=W0404

class ExpressionConditional( ExpressionChildrenHavingBase ):
    kind = "EXPRESSION_CONDITIONAL"

    named_children = ( "condition", "expression_yes", "expression_no" )

    def __init__( self, condition, yes_expression, no_expression, source_ref ):
        ExpressionChildrenHavingBase.__init__(
            self,
            values     = {
                "condition"      : condition,
                "expression_yes" : yes_expression,
                "expression_no"  : no_expression
            },
            source_ref = source_ref
        )

    def getBranches( self ):
        return ( self.getExpressionYes(), self.getExpressionNo() )

    getExpressionYes = ExpressionChildrenHavingBase.childGetter( "expression_yes" )
    getExpressionNo = ExpressionChildrenHavingBase.childGetter( "expression_no" )
    getCondition = ExpressionChildrenHavingBase.childGetter( "condition" )

    def computeExpression( self, constraint_collection ):
        # Children can tell all we need to know, pylint: disable=W0613
        condition = self.getCondition()

        # If the condition raises, we let that escape.
        if condition.willRaiseException( BaseException ):
            return condition, "new_raise", "Conditional expression raises in condition"

        # Decide this based on truth value of condition.
        truth_value = condition.getTruthValue()

        if truth_value is True:
            from .NodeMakingHelpers import wrapExpressionWithNodeSideEffects

            return (
                wrapExpressionWithNodeSideEffects(
                    new_node = self.getExpressionYes(),
                    old_node = condition
                ),
                "new_expression",
                "Conditional expression predicted to yes case"
            )
        elif truth_value is False:
            from .NodeMakingHelpers import wrapExpressionWithNodeSideEffects

            return (
                wrapExpressionWithNodeSideEffects(
                    new_node = self.getExpressionNo(),
                    old_node = condition
                ),
                "new_expression",
                "Conditional expression predicted to no case"
            )
        else:
            return self, None, None

    def mayHaveSideEffectsBool( self ):
        condition = self.getCondition()

        if condition.mayHaveSideEffectsBool():
            return True

        if self.getExpressionYes().mayHaveSideEffectsBool():
            return True

        if self.getExpressionNo().mayHaveSideEffectsBool():
            return True

        return False

    def mayProvideReference( self ):
        return self.getExpressionYes().mayProvideReference() or \
               self.getExpressionNo().mayProvideReference()


class StatementConditional( StatementChildrenHavingBase ):
    kind = "STATEMENT_CONDITIONAL"

    named_children = ( "condition", "yes_branch", "no_branch" )

    def __init__( self, condition, yes_branch, no_branch, source_ref ):
        StatementChildrenHavingBase.__init__(
            self,
            values     = {
                "condition"  : condition,
                "yes_branch" : yes_branch,
                "no_branch"  : no_branch
            },
            source_ref = source_ref
        )

    getCondition = StatementChildrenHavingBase.childGetter( "condition" )
    getBranchYes = StatementChildrenHavingBase.childGetter( "yes_branch" )
    setBranchYes = StatementChildrenHavingBase.childSetter( "yes_branch" )
    getBranchNo = StatementChildrenHavingBase.childGetter( "no_branch" )
    setBranchNo = StatementChildrenHavingBase.childSetter( "no_branch" )

    def isStatementAborting( self ):
        yes_branch = self.getBranchYes()

        if yes_branch is not None:
            if yes_branch.isStatementAborting():
                no_branch = self.getBranchNo()

                if no_branch is not None:
                    return no_branch.isStatementAborting()
                else:
                    return False
            else:
                return False
        else:
            return False


    def computeStatement( self, constraint_collection ):
        # This is rather complex stuff, pylint: disable=R0912

        # Query the truth value before the expression is evaluated, once it is
        # evaluated in onExpression, it may change. TODO: That is non-sense of
        # course, will be stable.
        condition = self.getCondition()
        truth_value = condition.getTruthValue()

        constraint_collection.onExpression( condition )
        condition = self.getCondition()

        from nuitka.optimizations.ConstraintCollections import ConstraintCollectionBranch

        # TODO: We now know that condition evaluates to true for the yes branch
        # and to not true for no branch

        yes_branch = self.getBranchYes()

        # Handle branches that became empty behind our back
        if yes_branch is not None:
            if not yes_branch.getStatements():
                yes_branch = None

        if yes_branch is not None:
            branch_yes_collection = ConstraintCollectionBranch( constraint_collection )

            branch_yes_collection.process( yes_branch )

            # May have just gone away, so fetch it again.
            yes_branch = self.getBranchYes()

            # If it's aborting, it doesn't contribute to merging.
            if yes_branch is None or yes_branch.isStatementAborting():
                branch_yes_collection = None
        else:
            branch_yes_collection = None

        no_branch = self.getBranchNo()

        if no_branch is not None:
            if not no_branch.getStatements():
                no_branch = None

        if no_branch is not None:
            branch_no_collection = ConstraintCollectionBranch( constraint_collection )

            branch_no_collection.process( no_branch )

            # May have just gone away, so fetch it again.
            no_branch = self.getBranchNo()

            # If it's aborting, it doesn't contribute to merging.
            if no_branch is None or no_branch.isStatementAborting():
                branch_yes_collection = None
        else:
            branch_no_collection = None

        # Merge into parent constraint collection.
        constraint_collection.mergeBranches(
            branch_yes_collection,
            branch_no_collection
        )

        if branch_yes_collection is not None and \
           branch_no_collection is not None:
            # TODO: Merging should be done by method.
            constraint_collection.variables = constraint_collection.mergeBranchVariables(
                branch_yes_collection.variables,
                branch_no_collection.variables
            )
        elif branch_yes_collection is not None:
            constraint_collection.mergeBranch( branch_yes_collection )
        elif branch_no_collection is not None:
            constraint_collection.mergeBranch( branch_no_collection )

        if yes_branch is None and no_branch is None:
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            # With both branches eliminated, the condition remains as a side
            # effect.
            result = makeStatementExpressionOnlyReplacementNode(
                expression = condition,
                node       = self
            )

            return result, "new_statements", """\
Both branches have no effect, reduced to evaluate condition."""

        if yes_branch is None:
            # Would be eliminated already, if there wasn't any "no" branch either.
            assert no_branch is not None

            from .OperatorNodes import ExpressionOperationNOT

            new_statement = StatementConditional(
                condition = ExpressionOperationNOT(
                    operand    = condition,
                    source_ref = condition.getSourceReference()
                ),
                yes_branch = no_branch,
                no_branch  = None,
                source_ref = self.getSourceReference()
            )

            return new_statement, "new_statements", """\
Empty 'yes' branch for condition was replaced with inverted condition check."""

        # Note: Checking the condition late, so that the surviving branches got
        # processed already. Returning without doing that, will corrupte the SSA
        # results.
        if truth_value is not None:
            from .NodeMakingHelpers import wrapStatementWithSideEffects

            if truth_value is True:
                choice = "true"

                new_statement = self.getBranchYes()
            else:
                choice = "false"

                new_statement = self.getBranchNo()

            new_statement = wrapStatementWithSideEffects(
                new_node   = new_statement,
                old_node   = condition,
                allow_none = True # surviving branch may empty
            )

            return new_statement, "new_statements", """\
Condition for branch was predicted to be always %s.""" % choice
        elif condition.willRaiseException( BaseException ):
            from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode

            result = makeStatementExpressionOnlyReplacementNode(
                expression = condition,
                node       = self
            )

            return result, "new_raise", """\
Conditional statements already raises implicitely in condition, removing \
branches"""

        return self, None, None
